use binrw::{BinRead, BinWrite};
use modular_bitfield::{
    bitfield,
    specifiers::{B1, B14},
};

use crate::{
    field_of::FieldOf,
    types::{Field, ObjectType, Point, WString, F64, U16, U32},
};

use self::{
    circle::Circle, ellipse::Ellipse, hatch::Hatch, line::Lines, polygon::Polygon,
    rectangle::Rectangle,
};

pub mod circle;
pub mod ellipse;
pub mod hatch;
pub mod line;
pub mod polygon;
pub mod rectangle;

#[bitfield(bits = 16)]
#[derive(BinRead, BinWrite, Debug, Copy, Clone)]
#[br(map = Self::from_bytes)]
#[bw(map = |&x| Self::into_bytes(x))]
pub struct ObjectFlags {
    pub disabled: B1,
    pub aspect_ratio_unlocked: B1,
    #[skip]
    __: B14,
}

/// Core properties defined for all object types
#[derive(BinRead, BinWrite, Debug)]
#[brw(magic(17u32))]
pub struct ObjectCore {
    pub pen: U32,
    pub obj_type: FieldOf<ObjectType>,
    pub flags: FieldOf<ObjectFlags>,
    pub name: WString,
    pub count: U32,
    _unknown_1: Field,
    pub io_control_enable_mask: U16,
    pub io_control_disable_mask: U16,
    _unknown_2: [Field; 5],
    pub origin: FieldOf<Point>,
    pub z: F64,
    pub a: F64,
    _unknown_3: Field,
}

#[derive(BinRead, BinWrite, Debug)]
pub enum Object {
    #[brw(magic = 1u32)]
    Curve(Lines),
    #[brw(magic = 2u32)]
    Point(Lines),
    #[brw(magic = 3u32)]
    Rectangle(Rectangle),
    #[brw(magic = 4u32)]
    Circle(Circle),
    #[brw(magic = 5u32)]
    Ellipse(Ellipse),
    #[brw(magic = 6u32)]
    Polygon(Polygon),
    #[brw(magic = 32u32)]
    Hatch(Hatch),
}
