use binrw::{BinRead, BinWrite, FilePtr64};

use crate::{
    field_of::FieldOf,
    types::{Field, Rgba, WString, WobbleType, F64, U32},
};

#[derive(BinRead, Debug)]
pub struct PenHeader {
    pub pen_count: u32,
    #[br(args {
        inner: binrw::args! {
            pen_count
        }
    })]
    pub data: FilePtr64<Pens>,
}

// Manually implement BinWrite as FilePtr does not support serialization
// See: https://github.com/jam1garner/binrw/issues/4
impl BinWrite for PenHeader {
    type Args<'a> = ();

    fn write_options<W: std::io::prelude::Write + std::io::prelude::Seek>(
        &self,
        writer: &mut W,
        endian: binrw::Endian,
        args: Self::Args<'_>,
    ) -> binrw::prelude::BinResult<()> {
        let pen_count: u32 = self.data.pens.len().try_into().unwrap();
        pen_count.write_options(writer, endian, args)?;

        // Write address of data, which is placed after this field
        let data_offset: u64 = writer.stream_position().unwrap() + 8;
        data_offset.write_options(writer, endian, args)?;

        self.data.pens.write_options(writer, endian, args)
    }
}

#[derive(BinRead, BinWrite, Debug)]
#[br(import {pen_count: u32})]
pub struct Pens {
    #[br(count = pen_count)]
    pub pens: Vec<Pen>,
}

#[derive(BinRead, BinWrite, Debug)]
#[brw(magic(236u32))] // Number of fields within this struct
pub struct Pen {
    pub color: FieldOf<Rgba>,
    pub name: WString,
    pub disabled: U32,
    pub use_default: U32,
    pub loop_count: U32,
    pub speed: F64, // Changes with wobble relative speed
    pub power: F64,
    pub frequency: U32,
    _unknown_1: Field,
    pub start_tc: U32,
    pub end_tc: U32,
    pub polygon_tc: U32,
    pub jump_speed: F64,
    _unknown_2: [Field; 10],
    pub laser_off_tc: U32,
    pub wave: U32, // Only available if continue_mode is false
    _unknown_3: Field,
    pub wobble_enable: U32,
    pub wobble_diameter: F64,
    pub wobble_distance: F64,
    _unknown_4: [Field; 8],
    pub min_jump_tc: U32,
    pub max_jump_tc: U32,
    pub jump_limit: F64,
    _unknown_5: [Field; 2],
    pub frequency_2: F64,
    _unknown_6: [Field; 152],
    pub wobble_type: FieldOf<WobbleType>,
    pub continue_mode: U32,
    _unknown_7: [Field; 12],
    pub wobble_diameter_2: F64, // Only with wobble type ellipse
    _unknown_8: [Field; 26],
}
